/*************************************************************                  
 *               IR PAD MFAD V2     F.MILLER 5/16
 ************************************************************* 
 * This sketch sends LocoNet commands (F0 - F8) based on received IR Pad commands 
 * IR Pad (0-8) keys cycle fcts ON/OFF of currently accessed LN address (1000, 1001)
 *        [>||] select address 1001
 *        [|<<] select address 1000
 * First time selection of either LN address will also ensure all functions are OFF
 * Detected LN error conditions will blink LEDs.  Best to reset to recover.
 * This sketch makes use of the LocoNet library developed by Alex Shepherd
 * Arduino Pin assignments:
 *  PIN 8 - LocoNet RX
 *  PIN 6 - Loconet TX
 *  PIN 14(A0) for IR Receive
 *  PIN 15(A1) for LN#1000 selected indicator
 *  PIN 16(A2) for LN#1001 selected indicator
 *  PIN 17(A3) for LN error signal
 *************************************************************/

#include <LocoNet.h>
#define RECALL_STACK_LEN  6
#define AJS_RECALL_STACK  { 38, 44, 53, 99, 191, 192 }
typedef struct {
  word  RecallStack[ RECALL_STACK_LEN ] ;
  word  LastAddress ;
  byte  LastSlot ;
} 
TH_EE_CONFIG ;
TH_EE_CONFIG    eeConfig __attribute__((section(".eeprom")))
  = { AJS_RECALL_STACK, 0, 0xFF };
LocoNetThrottleClass  Throttle ;
word LocoAddr ;
byte RecallIndex ;
LnBuf LnRxBuffer ;
lnMsg *RxPacket ;
uint32_t LastThrottleTimerTick;
long lastTime = 0;
long lndelaytime = 1000;        // pauses to allow LocoNet time to work
long tickTime = 500;            // Action clock ticks TickTime
int curAction;
word curAddr;                   // will be null (at start up) 1000 or 1001 
int curAddrIX;                  // will be 0 for LN1000 or 1 for 1001
int nxtFctOff;
#define ln1000LED A1            // LEDs to indicate currently acquired LN address
#define ln1001LED A2
boolean ln1000First = true;     // used to force setting Functions OFF at startup
boolean ln1001First = true;

#include <IRremote.h>
int RECV_PIN = 14;
IRrecv irrecv(RECV_PIN);
decode_results results;
int FctStates[2][9] = {
  {0,0,0,0,0,0,0,0,0},  //current function state F0-F8 for ln1000
  {0,0,0,0,0,0,0,0,0}   //current function state F0-F8 for ln1001
};

int onFctValue[9] = {16,1,2,4,8,1,2,4,8};    //these are the values returned when ON (0 for OFF)
int offFctValue = 0;
//================================= SETUP =======================================
void setup()  {
  pinMode(ln1000LED, OUTPUT);               //establish LN address LED indicators...
  pinMode(ln1001LED, OUTPUT);
  digitalWrite(ln1000LED, LOW);             //and initialize at OFF
  digitalWrite(ln1001LED, LOW);  
  LocoNet.init();                           //initialize the LocoNet interface
  Throttle.init(0, 0, 9999);
  Throttle.releaseAddress();                //release any LN addresses in LN library
  curAddr = 0;
  curAddrIX = 0;
  curAction = 0;
  //Serial.begin(57600);
  //Serial.println("START");
  irrecv.enableIRIn();                      // Start the IR receiver
  delay(1000);
}

//================================= LOOP =================================================
void loop() {
    RxPacket = LocoNet.receive(); // Check for any received LocoNet packets
  if( RxPacket ) {
    if( !LocoNet.processSwitchSensorMessage(RxPacket) ) 
      Throttle.processMessage(RxPacket) ; 
    }  
    if ((millis() > lastTime + tickTime)) {  // OK to do another action
      lastTime = millis();

    if (curAction != 0) {
    //Serial.print("doAction ");
    //Serial.println(curAction);
    doAction();
    } else {    //don't look for IR if doing a Loconet sequence  
    if (irrecv.decode(&results)) {
      //Serial.println(results.value, HEX);
      switch (results.value){
        case 0xFF6897:        //key 0
          doFcts(curAddrIX,0);
          break;
        case 0xFF30CF:        //key 1
          doFcts(curAddrIX,1);
          break;
        case 0xFF18E7:        //key 2
          doFcts(curAddrIX,2);
          break;
        case 0xFF7A85:        //key 3
          doFcts(curAddrIX,3);
          break;
        case 0xFF10EF:        //key 4
          doFcts(curAddrIX,4);
          break;
        case 0xFF38C7:        //key 5
          doFcts(curAddrIX,5);
          break;
        case 0xFF5AA5:        //key 6
          doFcts(curAddrIX,6);
          break;
        case 0xFF42BD:        //key 7
          doFcts(curAddrIX,7);
          break;
        case 0xFF4AB5:        //key 8
          doFcts(curAddrIX,8);
          break;
        case 0xFFC23D:      //key ">>|-" means set address to 1001;
          LocoAddr = 1001;
          curAddrIX = 1;
          if (LocoAddr != curAddr) {      //means new address assignment
            //Serial.print("Addr:");
            //Serial.println(LocoAddr);
            curAddr = LocoAddr;
            curAction = 1;
          }                               //... otherwise don't try to acquire again
          break;
        case 0xFF02FD:      //key "<<|-" means set address to 1000);
          LocoAddr = 1000;
          curAddrIX = 0;
          if (LocoAddr != curAddr) {      //means new address assignment
            //Serial.print("Addr:");
            //Serial.println(LocoAddr);
            curAddr = LocoAddr;
            curAction = 1;
          }                               //... otherwise don't try to acquire again
          break;
      }
      irrecv.resume(); // Receive the next value
    }
    delay(100);
    }
    if(isTime(&LastThrottleTimerTick, 100)) {
      Throttle.process100msActions() ; 
    }
  }
}

//=============================== SUPPORT ROUTINES ======================================

void doFcts(int LN,int F) {                //send applicable function commands
  //Serial.print("F");
  //Serial.print(F);
  if (!FctStates[LN][F]) {
    FctStates[LN][F] = onFctValue[F];
    //Serial.print("-ON for ");
  }else{
    FctStates[LN][F] = offFctValue;
    //Serial.print("-OFF for ");
  }
  Throttle.setFunction(F,FctStates[LN][F]);
  //Serial.println(LN);
}
//============================================================
void doAction() {                   //step through each set of commands per cycle in Loop
  switch (curAction) {              //gives opportunity for LocoNet library to process
    case 1:
      Throttle.releaseAddress();
      //Serial.println("Release");
      curAction++;
      break;    
    case 2:       //release address
      digitalWrite(ln1000LED, LOW);     //turn OFF both LN Address indicator LEDs
      digitalWrite(ln1001LED, LOW);
      //Serial.print("Freeing ");
      //Serial.println(curAddr);
      Throttle.freeAddress(curAddr);
      curAction++;
      break;
    case 3:       //locoNet activity delay
      //Serial.println("2-lag");
      delay(lndelaytime);
      curAction++;
      break;
    case 4:       // acquire new address
      //Serial.print("3-Acquire ");
      //Serial.println(curAddr);
      Throttle.setAddress(curAddr);
      curAction++;
      break;
     
    case 5:       //locoNet activity delay
      //Serial.println("5-lag");
      //delay(lndelaytime);
   
      if (curAddr == 1000) {
        if (ln1000First) {
          //Serial.println("1000 first time");
          nxtFctOff = 9;
          for (int F = 0;F < 9;F++)  {
                if (Throttle.getFunction(F)==onFctValue[F]) {       //send commands to LN
                  //delay(lndelaytime);
                  nxtFctOff=F;
                  //Serial.print(nxtFctOff);
                  //Serial.println ("still ON");
                }
          }
          if (nxtFctOff == 9) {     //see if captured any OFFs
            curAction = 0;
            //Serial.println("All Fct Off");
            ln1000First = false;
            digitalWrite(ln1000LED, HIGH);     //turn ON LN Address 1000 indicator LED
          } else {
            curAction++;          //found some so revisit
          }
        }else{
          curAction = 0;
          digitalWrite(ln1000LED, HIGH);     //turn ON LN Address 1000 indicator LED
        }
        
        //Serial.println(curAction);
        
        } else {                    // curAddr = 1001
           if (ln1001First) {
            //Serial.println("1001 first time");
            nxtFctOff = 9;
            for (int F = 0;F < 9;F++)  {
                if (Throttle.getFunction(F)==onFctValue[F]) {       //send commands to LN
                  //delay(lndelaytime);
                  nxtFctOff=F;
                  //Serial.print(nxtFctOff);
                  //Serial.println ("still ON");
                }
            }
            if (nxtFctOff == 9) {     //see if captured any OFFs
              curAction = 0;
              //Serial.println("All Fct Off");
              ln1001First = false;
              digitalWrite(ln1001LED, HIGH);     //turn ON LN Address 1000 indicator LED
            } else {
            curAction++;          //found some so revisit
            }
          }else{                  //not first time for LN
          curAction = 0;
          digitalWrite(ln1001LED, HIGH);     //turn ON LN Address 1000 indicator LED
          }
      break;
    case 6:       //turn off functions
      //Serial.println("6-Fct set off");
      Throttle.setFunction(nxtFctOff,offFctValue);
      //Serial.print(nxtFctOff);
      //Serial.println(" OFF");
      curAction++;
      break;
     case 7:
      //Serial.println("7-check done");
      if (nxtFctOff==9){
        curAction = 0;
      }else{
        curAction = 5;        //return to check for more Fcts
      }
      break;
    }
  }   
}      

// ================= Time Keeper for LocoNet ===================

boolean isTime(unsigned long *timeMark, unsigned long timeInterval) {
    unsigned long timeNow = millis();
    if ( timeNow - *timeMark >= timeInterval) {
        *timeMark = timeNow;
        return true;
    }    
    return false;
}
// ============== LOCONET CALL BACK AND SUPPORT FUNCTIONS =========

void notifyThrottleError( uint8_t UserData, TH_ERROR Error ) {
  //Serial.print("Error ");
  //Serial.print(Error, DEC);
  //Serial.print(' ');
  //Serial.println(Throttle.getErrorStr(Error));
  for (int i = 0; i < 20; i++)  {           //alternate flash LEDs to show LN error condition
      digitalWrite(ln1000LED, HIGH);
      digitalWrite(ln1001LED, LOW);
      delay(500);
      digitalWrite(ln1001LED, HIGH);
      digitalWrite(ln1000LED, LOW);
      delay(500);      
  }
}
//==================================================================

